// File _3d.h -   __main3d.h
//////////////////////////////////////////////////////////////////////////////
#ifndef  __3D__
#define  __3D__

//////////////////////////////////////////////////////////////////////////////
#include "_MATH_RandomBJ.h"
#include "_COARRAY.h"

//////////////////////////////////////////////////////////////////////////////
#define OBJ_Conv_Deep     256

#define DirX         0
#define DirY         1
#define DirZ         2
#define RewX         3
#define RewY         4
#define RewZ         5
#define RandOr       7

#define DirRt        1
#define RewRt       -1

#define MaxDT   0.0001
#define MinDT  -0.0001

//////////////////////////////////////////////////////////////////////////////
struct tbox { tvect mn,mx; };
struct decart { tvect Ox,Oy,Oz; };
typedef tfloat arrdec[3][3];
struct tmap { int Vx,Vy,Vz; };

//////////////////////////////////////////////////////////////////////////////                
struct SelectSp { unsigned num;
                  SelectSp far *next;
                };

//////////////////////////////////////////////////////////////////////////////                
inline tmap MAPis(int vx, int vy, int vz)
{
  tmap map;
  map.Vx = vx; map.Vy = vy; map.Vz = vz;
  return map;
};
           
inline decart DECARTis(tvect ox, tvect oy, tvect oz)
{
  decart de;
  de.Ox = ox; de.Oy = oy; de.Oz = oz;
  return de;
};
    
//////////////////////////////////////////////////////////////////////////////
class BaseOBJ
{ public:
  unsigned ToOBJ,DirOBJ,NextConv;
  tfloat Circ,SCirc,Ccos,Csin,Scos,Ssin;
  int Direct,Rottate,MvVCent;
  tvect DirPoint,ACenter,VCenter,QCenter,Scale,EScale;
  decart Dec,SDec,QDec;
  tmap Map,SMap;
  BaseOBJ();
  ~BaseOBJ();
  void SetOrient(unsigned m);
  void SetScale(tfloat Xs, tfloat Ys, tfloat Zs);
  void SetEScale(tfloat Xs, tfloat Ys, tfloat Zs);
  void SetMap(int Xv, int Yv, int Zv);
  void SetTSMap(tmap SMp);
  void SetSMap(int Xv, int Yv, int Zv);
  void SetCirc(tfloat ug);
  void SetSCirc(tfloat ug);
  void SetTDecart(decart De);
  void SetDecart(tfloat Xx, tfloat Xy, tfloat Xz, tfloat Yx, tfloat Yy, tfloat Yz,  tfloat Zx, tfloat Zy, tfloat Zz);
  void SetVCenter(tfloat xc, tfloat yc, tfloat zc);
  void SetACenter(tfloat xc, tfloat yc, tfloat zc);
  void MoveACenter(tfloat xc, tfloat yc, tfloat zc);
  void MoveVCenter(tfloat xc, tfloat yc, tfloat zc);
  void SetToOBJ(unsigned t);
  void SetDirOBJ(unsigned d);
  void SetDirPoint(tfloat x, tfloat y, tfloat z);
  void SetDirect(int d);
  void SetNextConv(unsigned n);
  void Default();
  void SubRotate(int k, int vst, int vnd);
  void Rotate();
  void NoRotate();
  tvect Conv(tvect p);
  tvect ConvVect(tvect p);
  tvect QConv(tvect p);
  tvect QConvVect(tvect p);
  tvect DeConv(tvect p);
  tvect DeConvVect(tvect p);
  tvect DeQConv(tvect p);
  tvect DeQConvVect(tvect p);
  void CMove(tfloat x, tfloat y, tfloat z);
};

//////////////////////////////////////////////////////////////////////////////
extern int BuildQDec;
extern int OBJ_Camera;
extern BaseCOARRAY OBJ;
extern BaseOBJ *COBJ;
extern int COBJn;
extern WCHAR *COBJname;

void OBJ_SetCamera(unsigned n);
void SetBuildQDec(int m);
void ResetQDec();
void RestoreQDec();

//////////////////////////////////////////
void ConsDec(tvect& v, decart *d);
void GetRandVect(tvect ov, tfloat r1, tfloat r2, tfloat s, tvect *v);
void OrientIn(tvect& p, unsigned to, tvect *p1);
void OrientOut(tvect& p, unsigned to, tvect *p1);
void SphConv(tvect p, unsigned to, tfloat *ax, tfloat *ay);
void CylConv(tvect p, unsigned to, tfloat *ax, tfloat *dy);

void OBJ_Copy(unsigned sn, unsigned tn);
void COBJ_CMove(tfloat x, tfloat y, tfloat z);

//////////////////////////////////////////////////////////////////////////////
inline void COBJ_GotoNextConv(unsigned n) { COBJ->SetNextConv(n); OBJ.Create(n); }

//////////////////////////////////////////////////////////////////////////////
//       QDec
//        
//     ""
align16 extern decart QDEC;
align16 extern decart CmQDEC;

inline tvect QDECconv(tvect& p)
{ return tvect(QDEC.Ox.x * p.x + QDEC.Oy.x * p.y + QDEC.Oz.x * p.z,
                QDEC.Ox.y * p.x + QDEC.Oy.y * p.y + QDEC.Oz.y * p.z,
                QDEC.Ox.z * p.x + QDEC.Oy.z * p.y + QDEC.Oz.z * p.z);
}

//////////////////////////////////////////////////////////////////////////////
inline tvect QDECdeconv(tvect& p) { return tvect(p & QDEC.Ox, p & QDEC.Oy, p & QDEC.Oz); }

//////////////////////////////////////////////////////////////////////////////
inline tvect CmQDECconv(tvect& p)
{ 
  return tvect(CmQDEC.Ox.x * p.x + CmQDEC.Oy.x * p.y + CmQDEC.Oz.x * p.z,  
                CmQDEC.Ox.y * p.x + CmQDEC.Oy.y * p.y + CmQDEC.Oz.y * p.z,
                CmQDEC.Ox.z * p.x + CmQDEC.Oy.z * p.y + CmQDEC.Oz.z * p.z);
}

//////////////////////////////////////////////////////////////////////////////
inline tvect CmQDECdeconv(tvect& p)
{
  return tvect(p & CmQDEC.Ox, p & CmQDEC.Oy, p & CmQDEC.Oz);           
}

//////////////////////////////////////////////////////////////////////////////
extern int QCNVOBJ;
extern int QCNV_mvVCenter;
extern tvect QCNV_VCenter;
extern tvect QCNV_QCenter;

extern int CmQCNV_mvVCenter;
extern tvect CmQCNV_VCenter;
extern tvect CmQCNV_QCenter;

//////////////////////////////////////////////////////////////////////////////
inline void QCNVset(unsigned i, int cm = Yes) //    i   
{                                                // Cm -   .
  QCNV_mvVCenter = CmQCNV_mvVCenter = 0;
  QCNVOBJ = i;
  if (QCNVOBJ && OBJ.Current(i))
  { QDEC = COBJ->QDec;
    QCNV_mvVCenter = COBJ->MvVCent;
    QCNV_VCenter = COBJ->VCenter;
    QCNV_QCenter = COBJ->QCenter;
  }  
  if (OBJ_Camera)
  { OBJ.Current(OBJ_Camera);
    CmQDEC = COBJ->QDec; 
    CmQCNV_mvVCenter = COBJ->MvVCent;
    CmQCNV_VCenter = COBJ->VCenter;
    CmQCNV_QCenter = COBJ->QCenter;
  }  
  //CmQCNVpoint_Link(cm);                
}

//////////////////////////////////////////////////////////////////////////////
inline tvect QCNVpoint(tvect p) //      
{ 
  if (QCNVOBJ) { if (QCNV_mvVCenter) p += QCNV_VCenter; p = QDECconv(p) + QCNV_QCenter; }  
  if (OBJ_Camera) { p = CmQDECdeconv(p - CmQCNV_QCenter); if (CmQCNV_mvVCenter) p -= CmQCNV_VCenter; }
  return p;    
}

//////////////////////////////////////////////////////////////////////////////
inline tvect QCNVvect(tvect v)
{ 
  if (QCNVOBJ) v = QDECconv(v);
  if (OBJ_Camera) return CmQDECdeconv(v); else return v;
}

//////////////////////////////////////////////////////////////////////////////
inline tvect DeQCNVpoint(tvect p)
{ 
  if (OBJ_Camera) { if (CmQCNV_mvVCenter) p += CmQCNV_VCenter; p = CmQDECconv(p) + CmQCNV_QCenter; }
  if (QCNVOBJ) { p = QDECdeconv(p - QCNV_QCenter); if (QCNV_mvVCenter) p -= QCNV_VCenter; }  
  return p;
}

//////////////////////////////////////////////////////////////////////////////
inline tvect DeQCNVvect(tvect v) //     QDECdeconv 
{ 
  if (OBJ_Camera) v = CmQDECconv(v);
  if (QCNVOBJ) return QDECdeconv(v); else return v; 
};

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
inline tvect ConvDec(tvect& p1, decart& De)
{
  return tvect(De.Ox.x * p1.x + De.Oy.x * p1.y + De.Oz.x * p1.z,
                De.Ox.y * p1.x + De.Oy.y * p1.y + De.Oz.y * p1.z,
                De.Ox.z * p1.x + De.Oy.z * p1.y + De.Oz.z * p1.z);
}

//////////////////////////////////////////////////////////////////////////////
inline tvect DeConvDec(tvect& p, decart& De)
{
  return tvect(p & De.Ox, p & De.Oy, p & De.Oz);
}

//////////////////////////////////////////////////////////////////////////////
inline tvect COBJ_Conv(tvect p)
{
  if (BuildQDec) p = COBJ->QConv(p); else { p = COBJ->Conv(p); unsigned rc = COBJn;
                                             while (COBJ->NextConv) { OBJ.Current(COBJ->NextConv); p = COBJ->Conv(p); }
                                             OBJ.Current(rc);
                                           }
  return p;
}

//////////////////////////////////////////////////////////////////////////////
inline tvect COBJ_ConvVect(tvect p)
{
  if (BuildQDec) return COBJ->QConvVect(p); else { p = COBJ->ConvVect(p); unsigned rc = COBJn;
                                                   while (COBJ->NextConv) { OBJ.Current(COBJ->NextConv); p = COBJ->ConvVect(p); }
                                                   OBJ.Current(rc);
                                                   return p;
                                                 }
}

//////////////////////////////////////////////////////////////////////////////
extern int indxconv[OBJ_Conv_Deep];

inline tvect COBJ_DeConv(tvect p)
{
  unsigned rc, i = 0;
  if (BuildQDec) return COBJ->DeQConv(p);
  else { rc = indxconv[i] = COBJn; i++;
         while (COBJ->NextConv != 0) { OBJ.Current(COBJ->NextConv); indxconv[i] = COBJn; i++; }
         while (i) { i--; OBJ.Current(indxconv[i]); p = COBJ->DeConv(p); }
         OBJ.Current(rc);
         return p;
       }
}

//////////////////////////////////////////////////////////////////////////////
inline tvect COBJ_DeConvVect(tvect p)
{ 
  unsigned rc, i = 0;
  if (BuildQDec) return COBJ->DeQConvVect(p);
  else { rc = indxconv[i] = COBJn; i++;
         while (COBJ->NextConv != 0) { OBJ.Current(COBJ->NextConv); indxconv[i] = COBJn; i++; }
         while (i) { i--; OBJ.Current(indxconv[i]); p = COBJ->DeConvVect(p); }
         OBJ.Current(rc);
         return p;
       }
}

//////////////////////////////////////////////////////////////////////////////
inline tvect Camera_DeConv(tvect p)
{
  if (OBJ_Camera) { unsigned rc = COBJn; OBJ.Current(OBJ_Camera);
                    p = COBJ_Conv(p); OBJ.Current(rc);
                  }
  return p;
}

//////////////////////////////////////////////////////////////////////////////
inline tvect COBJ_Camera_DeConv(tvect p) { return COBJ_DeConv(Camera_DeConv(p)); }

//////////////////////////////////////////////////////////////////////////////
inline tvect Camera_ConvVect(tvect p)
{
  if (OBJ_Camera) { unsigned rc = COBJn; OBJ.Current(OBJ_Camera);
                    p = COBJ_DeConvVect(p); OBJ.Current(rc);
                  }
  return p;
};
//////////////////////////////////////////////////////////////////////////////
inline tvect Camera_COBJ_ConvVect(tvect p) { return Camera_ConvVect(COBJ_ConvVect(p)); };
//////////////////////////////////////////////////////////////////////////////
inline tvect Camera_Conv(tvect p)
{
  if (OBJ_Camera) { unsigned rc=COBJn; OBJ.Current(OBJ_Camera);
                    p = COBJ_DeConv(p); OBJ.Current(rc);
                  }
  return p;
};
//////////////////////////////////////////////////////////////////////////////
inline tvect Camera_COBJ_Conv(tvect p) { return Camera_Conv(COBJ_Conv(p)); }

///////////////////////////////////////////////////////////////////////////////
void COBJ_QDec();
void ConsDirDec(tvect& v, decart *d);
void COBJ_DirToPoint(tvect p);
void COBJ_DirTo(word n);
void COBJ_Direct();

void OBJs_Rotate();        
void OBJs_QDec();                
void OBJs_RotateQDec();
void OBJs_Direct();
void OBJs_DirectRotateQDec();
void OBJs_DirectQDec();

//////////////////////////////////////////////////////////////////////////////
inline tvect ReflVect(tvect& iv, tvect& nv)
{
  tfloat D = (iv & nv) * 2;
  if (D > -0.0000001) return iv; else return iv - nv * D;
}

//////////////////////////////////////////////////////////////////////////////
#endif
