// File _DRAW_2D_GDIplus.cpp

///////////////////////////////////////////////////////////////////////////////
#include "stdafx.h"

#include "_DRAW_2D_GDIplus.h"

///////////////////////////////////////////////////////////////////////////////
int GDInow = no;

wchar_t *             GDIbitmapRT = RT_BITMAP; 

ULONG_PTR           GDItoken;
GdiplusStartupInput GDIinput;
Bitmap*             GDIbitmap = NULL;
BitmapData*         GDIbitmapData = NULL;
dwRGB_arr*          GDIbits = NULL;
int                 GDIbitmapLineSize;
int                 GDIbitmapSize;
int                 GDIbitmapXres;
int                 GDIbitmapYres;
PixelFormat         GDIpixelFormat = PixelFormat32bppARGB;

//////////////////////////////////////////////////////////////////////////////
int GDIbitmapLock(Bitmap* bitmap, PixelFormat pf) 
{
  if (!bitmap) return 0;
  GDIbitmapData = new BitmapData;
  Rect rect(0, 0, bitmap->GetWidth(), bitmap->GetHeight());
  bitmap->LockBits(&rect, ImageLockModeRead, pf, GDIbitmapData);
  GDIbitmapLineSize = GDIbitmapData->Stride;
  GDIbitmapXres = GDIbitmapData->Width;
  GDIbitmapYres = GDIbitmapData->Height;
  GDIbitmapSize = GDIbitmapYres * GDIbitmapLineSize; 
  GDIbits = (dwRGB_arr *)GDIbitmapData->Scan0;
  return 1;
}

//////////////////////////////////////////////////////////////////////////////
void GDIbitmapUnlock()
{
  if (GDIbitmapData && GDIbitmap) { GDIbitmap->UnlockBits(GDIbitmapData);
                                    delete GDIbitmapData; GDIbitmapData = NULL; GDIbits = NULL;
                                  }
}

///////////////////////////////////////////////////////////////////////////////
Bitmap* BitmapFromResource(wchar_t * name)
{
  if (!hmCommonResource) InitCommonResource();
  HRSRC hrsrc = GDIfindImageResource(hmCommonResource, name);
  if (!hrsrc) return NULL;
  if (GDIbitmapRT != RT_BITMAP)
  { HGLOBAL hg1 = LoadResource(hmCommonResource, hrsrc);
    DWORD sz = SizeofResource(hmCommonResource, hrsrc);
    void* ptr1 = LockResource(hg1);
    HGLOBAL hg2 = GlobalAlloc(GMEM_FIXED, sz);
    CopyMemory(LPVOID(hg2), ptr1, sz);
    IStream *pStream;
    HRESULT hr = CreateStreamOnHGlobal(hg2, TRUE, &pStream);
    if (FAILED(hr)) return 0;
    Bitmap *image = Bitmap::FromStream(pStream);
    pStream->Release();
    return image;
  } else { return (new Bitmap(hmCommonResource, name)); }
}

///////////////////////////////////////////////////////////////////////////////
int GDIbitmapFromResource(wchar_t * name)
{
  GDIbitmapRelase();
  GDIbitmap = BitmapFromResource(name);
  return (GDIbitmap != NULL);
}

//////////////////////////////////////////////////////////////////////////////
int GDIbitmapFromFile(wchar_t * name)
{
  GDIbitmapRelase();
  if (!PathFileExists(name)) return 0; 
  GDIbitmap = new Bitmap(name);
  return (GDIbitmap != NULL);
}

//////////////////////////////////////////////////////////////////////////////
int GDIbitmapLoad(wchar_t * name, int fromRes) //    fromRes = Yes;
                                               //  -  
                                               //   -     
{
  GDIbitmapRelase();
  if (fromRes == -1) 
  { if (GDIfindImageResource(hmCommonResource, name)) return GDIbitmapFromResource(name);
    else return GDIbitmapFromFile(name);
  } else if (fromRes == 1) return GDIbitmapFromResource(name);
         else return GDIbitmapFromFile(name);
}


//////////////////////////////////////////////////////////////////////////////
void GDIbitmapRelase()
{
  if (GDIbitmap) { delete GDIbitmap; GDIbitmap = NULL; }
}

//////////////////////////////////////////////////////////////////////////////
tfloat GDIbitmapScaleX = 0;
tfloat GDIbitmapScaleY = 0;
tfloat GDIbitmapScaleXres;
tfloat GDIbitmapScaleYres;
int GDIbitmapResizeMode = modeSIZEFIT; 
tfloat GDIbitmapResizeAspect;
tfloat GDIbitmapResizeRAspect;
int GDIbitmapResizeForm;
int GDIbitmapClear = no;


int GDIbitmapResize(int xr, int yr)
{
  if (!GDIbitmap) return no;
  GDIbitmapResizeAspect = (tfloat)xr / yr;
  GDIbitmapResizeRAspect = 1; GDIbitmapResizeForm = formEQUAL; 
  int sxr = GDIbitmap->GetWidth(), syr = GDIbitmap->GetHeight(); 
  if ((sxr == xr) && (syr == yr)) return Yes; 
  Bitmap *tmp = GDIbitmap;
  if (GDIbitmapScaleX > 0.00001) { xr = GDIbitmapScaleX * sxr + 0.5;
                                   if (GDIbitmapScaleY > 0.00001) yr = GDIbitmapScaleY * syr + 0.5; else yr = GDIbitmapScaleX * syr + 0.5; 
                                 } else if (yr == -1) yr = syr * xr / sxr;
  GDIbitmap = BitmapResize(GDIbitmap, xr, yr);
  delete tmp;
  GDIbitmapScaleX = GDIbitmapScaleY = 0;
  GDIbitmapScaleXres = (tfloat)xr / sxr; 
  GDIbitmapScaleYres = (tfloat)yr / syr; 
  return Yes;
}

//////////////////////////////////////////////////////////////////////////////
Bitmap *BitmapResize(Bitmap *source, int xr, int yr)
{
  Bitmap *result = new Bitmap(xr, yr, source->GetPixelFormat());
  ImageAttributes attr;
  int w, sxr = source->GetWidth(), syr = source->GetHeight();
  tfloat kdest = (tfloat)xr / yr, k;
  GDIbitmapResizeAspect = (tfloat)sxr / syr;
  k = GDIbitmapResizeRAspect = GDIbitmapResizeAspect / kdest;
  if ((int)(kdest * 1000) == (int)(GDIbitmapResizeAspect * 1000))
  { GDIbitmapResizeMode = modeSIZEFIT; GDIbitmapResizeRAspect = 1; GDIbitmapResizeForm = formEQUAL;
  } else if (GDIbitmapResizeRAspect > 1.0001) 
         { GDIbitmapResizeForm = formWIDE; w = Yes;
         } else if (GDIbitmapResizeRAspect < 0.9999) { GDIbitmapResizeForm = formHI; w = no; };
  Graphics g(result);
  g.SetInterpolationMode(InterpolationModeHighQualityBicubic);
  g.SetPixelOffsetMode(PixelOffsetModeHighQuality);
  g.SetSmoothingMode(SmoothingModeHighQuality);
  if (GDIbitmapClear) g.Clear(VS_CColor);
  switch (GDIbitmapResizeMode)
  { case modeSIZEFIT: attr.SetWrapMode(WrapModeTileFlipXY);
                      g.DrawImage(source, Rect(0,0,xr,yr), 0, 0, sxr, syr, UnitPixel, &attr); break;
    case modeSIZEREAL: if (w) { w = ((int)(syr * k + 0.5) - syr) / 2;
                                g.DrawImage(source, Rect(0,0,xr,yr), 0, -w, sxr, syr + w * 2, UnitPixel, &attr);
                              } else { w = ((int)(sxr / k + 0.5) - sxr) / 2;
                                       g.DrawImage(source, Rect(0,0,xr,yr), -w, 0, sxr + w * 2, syr, UnitPixel, &attr);
                                     }; break;
    case modeSIZEOUT: attr.SetWrapMode(WrapModeTileFlipXY);
                      if (w) { w = ((int)(sxr - (int)(sxr / k + 0.5))) / 2;
                               g.DrawImage(source, Rect(0,0,xr,yr), w, 0, sxr - w * 2, syr, UnitPixel, &attr);
                             } else { w = ((int)(syr - (int)(syr * k + 0.5))) / 2;
                                      g.DrawImage(source, Rect(0,0,xr,yr), 0, w, sxr, syr - w * 2, UnitPixel, &attr);
                                    }                                        
  }
  return result;
}  

///////////////////////////////////////////////////////////////////////////////
// work with files
int GDIbitmapSAVEtype = bitmapPNG;
WCHAR *GDIbitmapSAVEdelimiter01 = L"_";
WCHAR *GDIbitmapSAVEdelimiter02 = L"_";
WCHAR *GDIbitmapSAVEpath = L"";
WCHAR *GDIbitmapSAVEname = L"img";
WCHAR *GDIbitmapSAVEext = L"png";
WCHAR GDIbitmapSAVEfullname[1024] = L"";
unsigned GDIbitmapSAVEincBase = 0;
unsigned GDIbitmapSAVEincOffs = 1;
unsigned GDIbitmapSAVEincBaseDigits = 2;
unsigned GDIbitmapSAVEincOffsDigits = 8;

///////////////////////////////////////////////////////////////////////////////
void GDIbitmapSAVEfullnameDo()
{
  wcscpy_s(GDIbitmapSAVEfullname,L"");
  wcscat(GDIbitmapSAVEfullname, GDIbitmapSAVEpath);
  wcscat(GDIbitmapSAVEfullname, GDIbitmapSAVEname);
  if (GDIbitmapSAVEincBaseDigits)
  { wcscat(GDIbitmapSAVEfullname, GDIbitmapSAVEdelimiter01);
    NumToStr(GDIbitmapSAVEincBase, GDIbitmapSAVEincBaseDigits); 
    wcscat(GDIbitmapSAVEfullname, StrResult);
  }  
  if (GDIbitmapSAVEincOffsDigits)
  { wcscat(GDIbitmapSAVEfullname, GDIbitmapSAVEdelimiter02);
    NumToStr(GDIbitmapSAVEincOffs, GDIbitmapSAVEincOffsDigits); 
    wcscat(GDIbitmapSAVEfullname, StrResult);
  }  
  wcscat(GDIbitmapSAVEfullname,L".");
  wcscat(GDIbitmapSAVEfullname,GDIbitmapSAVEext);
}

///////////////////////////////////////////////////////////////////////////////
int GetEncoderClsid(const WCHAR* format, CLSID* pClsid)
{
   unsigned  num = 0;          // number of image encoders
   unsigned  size = 0;         // size of the image encoder array in bytes
   ImageCodecInfo* pImageCodecInfo = NULL;
   GetImageEncodersSize(&num, &size);
   if (size == 0) return -1;  // Failure
   pImageCodecInfo = (ImageCodecInfo*)(malloc(size));
   if (pImageCodecInfo == NULL) return -1;  // Failure
   GetImageEncoders(num, size, pImageCodecInfo);
   for (unsigned j = 0; j < num; ++j)
   if (wcscmp(pImageCodecInfo[j].MimeType, format) == 0)
    { *pClsid = pImageCodecInfo[j].Clsid;
       free(pImageCodecInfo);
       return j;  // Success
    };    
   free(pImageCodecInfo);
   return -1;  // Failure
}

///////////////////////////////////////////////////////////////////////////////
int GDIgetImageEncoder(CLSID &Clsid)
{
  switch (GDIbitmapSAVEtype) 
  { case bitmapBMP: GDIbitmapSAVEext = L"bmp"; return GetEncoderClsid(L"image/bmp", &Clsid);
    case bitmapJPG: GDIbitmapSAVEext = L"jpg"; return GetEncoderClsid(L"image/jpeg", &Clsid);
    case bitmapGIF: GDIbitmapSAVEext = L"gif"; return GetEncoderClsid(L"image/gif", &Clsid);
    case bitmapTIF: GDIbitmapSAVEext = L"tif"; return GetEncoderClsid(L"image/tiff", &Clsid);
    case bitmapPNG: GDIbitmapSAVEext = L"png"; return GetEncoderClsid(L"image/png", &Clsid);
    default: GDIbitmapSAVEext = L"bmp"; return GetEncoderClsid(L"image/bmp", &Clsid);
  }
}

///////////////////////////////////////////////////////////////////////////////
int GDIbitmapSAVE(BYTE *bits, unsigned xr, unsigned yr, int loffs, PixelFormat pf)
{
  CLSID Clsid;
  Bitmap *Bmp;
  if (GDIgetImageEncoder(Clsid) == -1) return 0;
  if (!bits) { Bmp = new Bitmap(hImage, NULL); xr = VS_X; yr = VS_Y; if (loffs == -1) loffs = xr << 2;  
             } else { if (loffs == -1) loffs = xr << 2; 
                      Bmp = new Bitmap(xr, yr, loffs, pf, bits);
                    };
  GDIbitmapSAVEfullnameDo();
  Bmp->Save(GDIbitmapSAVEfullname, &Clsid, NULL);         
  delete Bmp; return 1;  
}

///////////////////////////////////////////////////////////////////////////////
int GDIbitmapSAVE(wchar_t *path, BYTE *bits, unsigned xr, unsigned yr, int loffs, PixelFormat pf)
{
  CLSID Clsid;
  Bitmap *Bmp;
  if (GDIgetImageEncoder(Clsid) == -1) return 0;
  if (!bits) { Bmp = new Bitmap(hImage, NULL); xr = VS_X; yr = VS_Y; if (loffs == -1) loffs = xr << 2;  
             } else { if (loffs == -1) loffs = xr << 2; 
                      Bmp = new Bitmap(xr, yr, loffs, pf, bits);
                    };
  Bmp->Save(path, &Clsid, NULL);         
  delete Bmp; return 1;  
}

////////////////////////////////////////////////////////////////////////////////
int GDIsshotXR = 0;
int GDIsshotYR = 0;

int GDIsshotSAVE(wchar_t *path1, wchar_t *path2, BYTE *bits, unsigned xr, unsigned yr, int loffs)
{
  CLSID Clsid;
  Bitmap *Bmp;
  if (GDIgetImageEncoder(Clsid) == -1) return 0;
  if (!bits) { Bmp = new Bitmap(hImage,NULL); xr = VS_X; yr = VS_Y; if (loffs == -1) loffs = xr << 2;  
             } else { if (loffs == -1) loffs = xr << 2; 
                      Bmp = new Bitmap(xr, yr, loffs, PixelFormat32bppRGB, bits);
                    }
  if (path1) { GDIbitmapSAVEpath = path1;
               GDIbitmapSAVEfullnameDo();
               Bmp->Save(GDIbitmapSAVEfullname, &Clsid ,NULL);
             }  
  if (path2) { if (GDIsshotXR) { xr = GDIsshotXR; yr = GDIsshotYR;
                               } else { xr /= 2; yr /= 2; }                                      
               GDIbitmapSAVEpath = path2;
               GDIbitmapSAVEfullnameDo();
               if ((xr == VS_X) && (yr == VS_Y)) 
                { Bmp->Save(GDIbitmapSAVEfullname, &Clsid, NULL);
                } else { Bitmap *BmpR = BitmapResize(Bmp, xr, yr);
                         BmpR->Save(GDIbitmapSAVEfullname, &Clsid, NULL);
                         delete BmpR;
                       }  
             } 
  delete Bmp; return 1;  
}

///////////////////////////////////////////////////////////////////////////////
void GDIbitmapSAVEincBaseDo() { GDIbitmapSAVEincBase++; GDIbitmapSAVEincOffs = 1; } 
void GDIbitmapSAVEincOffsDo() { GDIbitmapSAVEincOffs++; } 

///////////////////////////////////////////////////////////////////////////////