// File _MATH_Noise.cpp
//////////////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include "_MATH_Noise.h"

inline void Interp(int it, tfloat i, tfloat k, tfloat &ri, tfloat &rk)
{
  switch(it)
  { case interpLINEAR: ri = i; rk = 0; break;
    case interpHERMIT: ri = Hermit(i); rk = 0; break; 
    case interpCAVE: ri = i; rk = -Hump(i) * k; break;
    case interpHUMP: ri = i; rk = Hump(i) * k; break;
  };
};

//////////////////////////////////////////////////////////////////////////////
Noise1D :: Noise1D()
{
  DPtr = NULL;
  FPtr = NULL;
//  Make(10,10,10);
//  Generate();
};
//////////////////////////////////////////////////////////////////////////////
Noise1D :: ~Noise1D()
{
  FreeMem(FPtr);
  FreeMem(DPtr);
};

//////////////////////////////////////////////////////////////////////////////
void Noise1D :: Clear()
{
  memset(DPtr,0,size);
};
//////////////////////////////////////////////////////////////////////////////
void Noise1D :: Make(unsigned xs, tfloat ln)
{
  FreeMem(DPtr);
  Sec = xs; d = (xs+1); fd = d;
  size = d * sizeof(tfloat);
  LnSec = ln;
  DPtr= (tfloat_arr *)GetMem(size);
  Generate();
};
//////////////////////////////////////////////////////////////////////////////
void Noise1D :: Generate()
{
  for (int x = 0; x <= Sec; x++)(*DPtr)[x] = Random();
};
//////////////////////////////////////////////////////////////////////////////
void Noise1D :: GenerateNsfast(tfloat base, tfloat jitter, int t, tfloat tt)
{
  tfloat f;
  Secfast = LnSec * (Sec + 1) + 0.5;
  FreeMem(FPtr); FPtr = (tfloat_arr *)GetMem(Secfast * sizeof(tfloat));
  for (int x = 0; x < Secfast; x++)(*FPtr)[x] =  base - jitter + 2 * jitter * GetNs(x, t, tt);
};

//////////////////////////////////////////////////////////////////////////////
tfloat Noise1D :: GetNsfast(int x)
{ if mI(x) x = Secfast - abs(x) % Secfast;
  return (*FPtr)[x % Secfast];
};
//////////////////////////////////////////////////////////////////////////////
tfloat Noise1D :: GetNs(tfloat x, int t, tfloat tt)
{
  x /= LnSec;
  x = fmod(x, fd); if (x < 0) x = fd + x;
  int x1 = x; int x2 = x1 + 1; if (x2 == d) x2=0;
  x = x - x1;
  tfloat i1 = (*DPtr)[x1]; tfloat i2 = (*DPtr)[x2], k = 0;
  if (t) Interp(t, x, tt, x, k);
  return i1 + (i2 - i1) * x + k;
};

//////////////////////////////////////////////////////////////////////////////
tfloat Noise1D :: Noise(tfloat x, unsigned c, int t, tfloat tt)
{
  tfloat e = 1; tfloat d = GetNs(x, t);
  for (int i = 1; i <= c; i++) { x *= 1.7773; e *= 0.5; d += GetNs(x, t, tt) * e; };
  return d;
};
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
inline tfloat NS2Dget(tfloat_arr *ptr, unsigned pat, unsigned x, unsigned y) {return (*ptr)[pat * y + x];};
inline void NS2Dput(tfloat_arr *ptr, unsigned pat, unsigned x, unsigned y, tfloat t) {(*ptr)[pat * y + x] = t;};

//////////////////////////////////////////////////////////////////////////////
Noise2D :: Noise2D()
{
  DPtr=NULL;
//  Make(10,10,10);
//  Generate();
};
//////////////////////////////////////////////////////////////////////////////
Noise2D :: ~Noise2D()
{
  FreeMem(DPtr);
};

//////////////////////////////////////////////////////////////////////////////
void Noise2D :: Clear()
{
  memset(DPtr,0,size);
};
//////////////////////////////////////////////////////////////////////////////
void Noise2D :: Make(unsigned xs, unsigned ys, tfloat ln)
{
  FreeMem(DPtr);
  SecX = xs; dx = (xs+1); fdx = dx;
  SecY = ys; dy = (ys+1); fdy = dy;
  pat = dx;
  size = dy * dy * pat * sizeof(tfloat);
  LnSec = ln;
  DPtr= (tfloat_arr *)GetMem(size);
  Generate();
};
//////////////////////////////////////////////////////////////////////////////
void Noise2D :: Generate()
{
  unsigned x,y;
  for (x=0;x<=SecX;x++)
   for (y=0;y<=SecY;y++) NS2Dput(DPtr,pat,x,y,Random());
};

//////////////////////////////////////////////////////////////////////////////
tfloat Noise2D :: GetNs(tfloat x, tfloat y, int t, tfloat tt, int t2)
{
  x /= LnSec;
  y /= LnSec;
  x = fmod(x, fdx); if (x<0) x = fdx + x;
  y = fmod(y, fdy); if (y<0) y = fdy + y;
  int x1 = x; int x2 = x1 + 1;
  int y1 = y; int y2 = y1 + 1;
  x = x - x1;
  y = y - y1;
  int f = Yes;
  if (x2 == dx) {x2=0; f = No;};
  if (y2 == dy) {y2=0; f = No;};
  tfloat i1,i2,ii1,ii2,k = 0;
  if (f)
   { tfloat *ptr = &((*DPtr)[pat * y1 + x1]);
     i1 = *ptr; ptr++;
     i2 = *ptr; ptr += pat;
     ii2 = *ptr; ptr--;
     ii1 = *ptr;
   } else { i1 = NS2Dget(DPtr,pat,x1, y1); i2 = NS2Dget(DPtr,pat,x2, y1);
            ii1 = NS2Dget(DPtr,pat,x1, y2); ii2 = NS2Dget(DPtr,pat,x2, y2);
          };  
  if (t) Interp(t, y, tt, y, k);
  i1 = i1 + (ii1 - i1) * y + k;
  i2 = i2 + (ii2 - i2) * y + k;
  if (t2 == -1) t2 = t;        
  k = 0; if (t2) Interp(t2, x, tt, x, k);
  return i1 + (i2 - i1) * x + k;
};

//////////////////////////////////////////////////////////////////////////////

tfloat Noise2D :: Noise(tfloat x, tfloat y, unsigned c, int t, tfloat tt, int t2)
{
  tfloat e = 1; tfloat d = GetNs(x, y, t);
  for (int i = 1; i <= c; i++) { x *= 1.7773; y *= 1.7773; e *= 0.5; d += GetNs(x, y, t, tt, t2) * e; };
  return d;
};
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
inline tfloat NS3Dget(tfloat_arr *ptr, unsigned pat1, unsigned pat2, unsigned x, unsigned y, unsigned z) {return (*ptr)[pat1 * z + pat2 * y + x];};
inline void NS3Dput(tfloat_arr *ptr, unsigned pat1, unsigned pat2, unsigned x, unsigned y, unsigned z, tfloat t) {(*ptr)[pat1 * z + pat2 * y + x] = t;};

//////////////////////////////////////////////////////////////////////////////
Noise3D :: Noise3D()
{
  DPtr=NULL;
  Make(10,10,10,10);
};

//////////////////////////////////////////////////////////////////////////////
Noise3D :: ~Noise3D()
{
  FreeMem(DPtr);
};

//////////////////////////////////////////////////////////////////////////////
void Noise3D :: Clear()
{
  memset(DPtr,0,size);
};
//////////////////////////////////////////////////////////////////////////////
void Noise3D :: Make(unsigned xs, unsigned ys, unsigned zs, tfloat ln)
{
  FreeMem(DPtr);
  SecX = xs; dx = (xs+1); fdx = dx;
  SecY = ys; dy = (ys+1); fdy = dy;
  SecZ = zs; dz = (zs+1); fdz = dz;
  pat2 = dx;
  pat1 = dx * dy;
  size = dx * dy * dz * sizeof(tfloat);
  LnSec = ln;
  DPtr = (tfloat_arr *)GetMem(size);
  Generate();
};
//////////////////////////////////////////////////////////////////////////////
void Noise3D :: Generate()
{
  unsigned x,y,z;
  for (x = 0; x <= SecX; x++)
   for (y = 0; y <= SecY; y++)
    for (z = 0; z <= SecZ; z++) NS3Dput(DPtr,pat1,pat2,x,y,z,Random());
};

/////////////////////////////////////////////////////////////////////////////
tfloat Noise3D :: GetNs(tvect p, int t, tfloat tt)
{
  tfloat i1,i2,i3,i4,ii1,ii2,ii3,ii4, k = 0;
  int x1,y1,z1,x2,y2,z2;

  p.x /= LnSec;
  p.y /= LnSec;
  p.z /= LnSec;
  p.x = fmod(p.x, fdx); if (p.x < 0) p.x = fdx + p.x;
  p.y = fmod(p.y, fdy); if (p.y < 0) p.y = fdy + p.y;
  p.z = fmod(p.z, fdz); if (p.z < 0) p.z = fdz + p.z;
  x1 = p.x; x2 = x1 + 1;
  y1 = p.y; y2 = y1 + 1;
  z1 = p.z; z2 = z1 + 1;
  p.x = p.x - x1;
  p.y = p.y - y1;
  p.z = p.z - z1;
  int f = Yes;
  if (x2 == dx) {x2=0; f = No;};
  if (y2 == dy) {y2=0; f = No;};
  if (z2 == dz) {z2=0; f = No;};
  if (f)
   { tfloat *ptr = &((*DPtr)[pat1*z1 + pat2 * y1 + x1]);
     i1 = *ptr; ptr++;
     i2 = *ptr; ptr += pat2;
     i4 = *ptr; ptr--;
     i3 = *ptr; ptr += pat1;
     ii3 = *ptr; ptr++;
     ii4 = *ptr; ptr -= pat2;
     ii2 = *ptr; ptr--;
     ii1 = *ptr;
   } else { i1 = NS3Dget(DPtr,pat1,pat2,x1,y1,z1);
            i2 = NS3Dget(DPtr,pat1,pat2,x2,y1,z1);
            i3 = NS3Dget(DPtr,pat1,pat2,x1,y2,z1);
            i4 = NS3Dget(DPtr,pat1,pat2,x2,y2,z1);
            ii1 = NS3Dget(DPtr,pat1,pat2,x1,y1,z2);
            ii2 = NS3Dget(DPtr,pat1,pat2,x2,y1,z2);
            ii3 = NS3Dget(DPtr,pat1,pat2,x1,y2,z2);
            ii4 = NS3Dget(DPtr,pat1,pat2,x2,y2,z2);
          };
  if (t) Interp(t, p.z, tt, p.z, k);        
  i1 = i1 + (ii1 - i1) * p.z + k;
  i2 = i2 + (ii2 - i2) * p.z + k;
  i3 = i3 + (ii3 - i3) * p.z + k;
  i4 = i4 + (ii4 - i4) * p.z + k;
  k = 0; if (t) Interp(t, p.x, tt, p.x, k);
  i1 = i1 + (i2 - i1) * p.x + k;
  i3 = i3 + (i4 - i3) * p.x + k;
  k = 0; if (t) Interp(t, p.y, tt, p.y, k);
  return i1 + (i3 - i1) * p.y + k;
};
//////////////////////////////////////////////////////////////////////////////
tfloat Noise3D :: Noise(tvect p, unsigned c, int t, tfloat tt)
{
  tfloat e = 1; tfloat d = GetNs(p,t);
  for (int i = 1; i <= c; i++) { p *= 1.7773; e *= 0.5; d += GetNs(p, t, tt) * e; };
  return d;
};
//////////////////////////////////////////////////////////////////////////////
