#ifndef _MATH_Vector__
#define _MATH_Vector__
#include "_MATH.h"
////////////////////////////////////////////////////////////////////////////////
extern tfloat ModVect;

////////////////////////////////////////////////////////////////////////////////
class tvect
{ public:
  tfloat x, y, z;
  tvect() {};
  tvect(tfloat v) { x = y = z = v; };
  tvect(const tvect& v) { x = v.x; y = v.y; z = v.z; };
  tvect(tfloat vx, tfloat vy, tfloat vz) { x = vx; y = vy; z = vz; };
  tvect& operator = (const tvect& v) {x = v.x; y = v.y; z = v.z; return *this; };
  tvect& operator = (tfloat v) { x = y = z = v; return *this; };
  tvect operator - () const { return tvect(-x, -y, -z); };
  tvect& operator += (const tvect& v) { x += v.x; y += v.y; z += v.z; return *this; };
  tvect& operator -= (const tvect& v) { x -= v.x; y -= v.y; z -= v.z; return *this; };
  tvect& operator *= (const tvect& v) { x *= v.x; y *= v.y; z *= v.z; return *this; };
  tvect& operator *= (tfloat d) { x *= d; y *= d; z *= d; return *this; };
  tvect& operator /= (const tvect& v) { x /= v.x; y /= v.y; z /= v.z; return *this; };
  tvect& operator /= (tfloat d) { x /= d; y /= d; z /= d; return *this; };
  friend tvect operator + (const tvect& v1, const tvect& v2);
  friend tvect operator - (const tvect& v1, const tvect& v2);
  friend tvect operator * (const tvect& v1, const tvect& v2);
  friend tvect operator * (tfloat d, const tvect& v2);
  friend tvect operator * (const tvect& v1, tfloat d);
  friend tvect operator / (const tvect& v1, const tvect& v2);
  friend tvect operator / (const tvect& v1, tfloat d);
  friend tfloat operator &(const tvect& v1, const tvect& v2);
  friend tvect operator ^(const tvect& v1, const tvect& v2);
  tfloat operator !() {return ModVect = sqrt(x*x + y*y + z*z);};
//  tvect operator | () {ModVect = sqrt(x*x+y*y+z*z); if (ModVect > 1e-16) {x/=ModVect; y/=ModVect; z/=ModVect;}; return *this;};
  tfloat& operator [](int n) {return * (&x + n);};
  int IsZero() {return (x<1e-16 && y<1e-16 && z<1e-16);};
  void Rew() {x=-x; y=-y; z=-z;};
  void Norm() {ModVect = sqrt(x*x+y*y+z*z); if (ModVect > 1e-16) {x/=ModVect; y/=ModVect; z/=ModVect;};};
  void wcNorm() {ModVect = sqrt(x*x+y*y+z*z); x/=ModVect; y/=ModVect; z/=ModVect;}; //     
  void Vect(tvect& v1, tvect& v2) {x=v2.x-v1.x; y=v2.y-v1.y; z=v2.z-v1.z;};
  void Rand(tfloat jtr = 1) {x = fRandom(-jtr,jtr); y = fRandom(-jtr,jtr); z = fRandom(-jtr,jtr);};
  void NormRand() {Rand(); wcNorm();};
  void SQRand(tfloat jtr = 1) { Rand();
                                tfloat ax = fabs(x), ay = fabs(y), az = fabs(z), d;
                                if ((ax > ay) && (ax > az)) d = jtr / ax; else if (ay > az) d = jtr / ay; else d = jtr / az;
                                x *= d; y *= d; z *= d;
                              };
  void NormToSQRand(tfloat t = 0, tfloat jtr = 1) //  -     t = 0..1  
       { Rand();
         if (t == 0) {wcNorm(); return;}
         tfloat ax = fabs(x), ay = fabs(y), az = fabs(z), d;
         if ((ax > ay) && (ax > az)) d = jtr / ax; else if (ay > az) d = jtr / ay; else d = jtr / az;
         if (t == 1) {x *= d; y *= d; z *= d; return;};
         ax = x * d; ay = y * d; az = z * d;
         wcNorm();
         x = x + (ax - x) * t; y = y + (ay - y) * t; z = z + (az - z) * t;
       };
  //int NEmore(tfloat d) {return (abs(x) > d) | (abs(y) > d) | (abs(z) > d);} //   d - 
  tfloat SQAbs() {return (x*x+y*y+z*z);};
};
////////////////////////////////////////////////////////////////////////////////
inline tvect operator + (const tvect& v1, const tvect& v2)
{
  return tvect(v1.x + v2.x, v1.y + v2.y, v1.z + v2.z);
};

inline tvect operator - (const tvect& v1, const tvect& v2)
{
  return tvect(v1.x - v2.x, v1.y - v2.y, v1.z - v2.z);
};

inline tvect operator * (const tvect& v1, const tvect& v2)
{
  return tvect(v1.x * v2.x, v1.y * v2.y, v1.z * v2.z);
};

inline tvect operator * (tfloat d, const tvect& v2)
{
  return tvect(d * v2.x, d * v2.y, d * v2.z);
};

inline tvect operator * (const tvect& v1, tfloat d)
{
  return tvect(v1.x * d, v1.y * d, v1.z * d);
};

inline tvect operator / (const tvect& v1, const tvect& v2)
{
  return tvect(v1.x / v2.x, v1.y / v2.y, v1.z / v2.z);
};

inline tvect operator / (const tvect& v1, tfloat d)
{
  return tvect(v1.x / d, v1.y / d, v1.z / d);
};

inline tfloat operator &(const tvect& v1, const tvect& v2)
{
  return (tfloat)(v1.x*v2.x + v1.y*v2.y + v1.z*v2.z);
};

inline tvect operator ^(const tvect& v1, const tvect& v2)
{
  return tvect(v1.y*v2.z-v1.z*v2.y, v1.z*v2.x-v1.x*v2.z, v1.x*v2.y-v1.y*v2.x);
};
////////////////////////////////////////////////////////////////////////////////
inline tvect Norm(tvect &v)
{
  ModVect = !v;
  if (ModVect > 1e-16) { return (v / ModVect);
                       } else return v;
};
inline int LDir(tvect &v1, tvect &v2)
{
  return ((v1 & v2)>0);
};
////////////////////////////////////////////////////////////////////////////////
// ,  
inline tvect VectBlend(tvect& v1, tvect& v2, tfloat b)  //   v1   v2,  b = 0..1
{
  return tvect(v1.x + (v2.x - v1.x) * b, v1.y + (v2.y - v1.y) * b, v1.z + (v2.z - v1.z) * b);
};

////////////////////////////////////////////////////////////////////////////////
class trgb
{ public:
  tfloat r, g, b;
  trgb() {};
  trgb(tfloat v) { r = g = b = v; };
  trgb(const trgb& v) {r = v.r; g = v.g; b = v.b;};
  trgb(tfloat vr, tfloat vg, tfloat vb) {r = vr; g = vg; b = vb;};
  trgb& operator = (const trgb& v) {r = v.r; g = v.g; b = v.b; return *this;};
  trgb& operator = (tfloat v) {r = g = b = v; return *this;};
  trgb operator - () const { return trgb(-r, -g, -b); };
  trgb& operator += (const trgb& v) { r += v.r; g += v.g; b += v.b; return *this; };
  trgb& operator -= (const trgb& v) { r -= v.r; g -= v.g; b -= v.b; return *this; };
  trgb& operator *= (const trgb& v) { r *= v.r; g *= v.g; b *= v.b; return *this; };
  trgb& operator *= (tfloat d) { r *= d; g *=d; b*=d; return *this; };
  trgb& operator /= (const trgb& v) { r /= v.r; g /= v.g; b /= v.b; return *this; };
  trgb& operator /= (tfloat d) { r /= d; g /= d; b /= d; return *this; };
  friend trgb operator + (const trgb& v1, const trgb& v2);
  friend trgb operator - (const trgb& v1, const trgb& v2);
  friend trgb operator * (const trgb& v1, const trgb& v2);
  friend trgb operator * (tfloat d, const trgb& v2);
  friend trgb operator * (const trgb& v1, tfloat d);
  friend trgb operator / (const trgb& v1, const trgb& v2);
  friend trgb operator / (const trgb& v1, tfloat d);
  tfloat& operator [](int n) {return * (&r + n);};
  int IsZero() {return (r < 1e-16 && g < 1e-16 && b < 1e-16);};
  void Clip() { if (r > 1) r = 1; else if (r < 0) r = 0;
                if (g > 1) g = 1; else if (g < 0) g = 0;
                if (b > 1) b = 1; else if (b < 0) b = 0;
              };
  void Blend(trgb v, tfloat t) { r += (v.r - r) * t; g += (v.g - g) * t; b += (v.b - b) * t; };            
  void Blend(tfloat vr, tfloat vg, tfloat vb, tfloat t) { r += (vr - r) * t; g += (vg - g) * t; b += (vb - b) * t; };            
  tfloat Midle() {return ((r + g + b) / 3);};
  tfloat ClipMidle()
     { tfloat t= (r + g + b) / 3;
       if (t>1) { return 1;
                } else return t;
     };
};
////////////////////////////////////////////////////////////////////////////////
inline trgb operator + (const trgb& v1, const trgb& v2)
{
  return trgb(v1.r + v2.r, v1.g + v2.g, v1.b + v2.b);
};

inline trgb operator - (const trgb& v1, const trgb& v2)
{
  return trgb(v1.r - v2.r, v1.g - v2.g, v1.b - v2.b);
};

inline trgb operator * (const trgb& v1, const trgb& v2)
{
  return trgb(v1.r * v2.r, v1.g * v2.g, v1.b * v2.b);
};

inline trgb operator * (tfloat d, const trgb& v2)
{
  return trgb(d * v2.r, d * v2.g, d * v2.b);
};

inline trgb operator * (const trgb& v1, tfloat d)
{
  return trgb(v1.r * d, v1.g * d, v1.b * d);
};

inline trgb operator / (const trgb& v1, const trgb& v2)
{
  return trgb(v1.r / v2.r, v1.g / v2.g, v1.b / v2.b);
};

inline trgb operator / (const trgb& v1, tfloat d)
{
  return trgb(v1.r / d, v1.g / d, v1.b / d);
};

////////////////////////////////////////////////////////////////////////////////
void SQRandVect();

////////////////////////////////////////////////////////////////////////////////
#endif
